# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""List of a file's metadata."""

from gi.repository import GObject, Gtk
from metadatacleaner import RESOURCE_PATH
from metadatacleaner.modules.metadata import MetadataList
from metadatacleaner.ui.metadatadetailsrow import MetadataDetailsRow


@Gtk.Template(resource_path=RESOURCE_PATH + "/ui/MetadataDetails.ui")
class MetadataDetails(Gtk.Box):
    """List of a file's metadata."""

    __gtype_name__ = "MetadataDetails"

    filename = GObject.Property(type=str)

    _metadata_list = MetadataList()

    _list: Gtk.ListBox = Gtk.Template.Child()

    @GObject.Property(type=MetadataList)
    def metadata_list(self) -> MetadataList:
        """Metadata List.

        Returns:
            MetadataList: A list of Metadata.
        """
        return self._metadata_list

    @metadata_list.setter  # type: ignore
    def metadata_list(self, metadata_list):
        self._metadata_list = metadata_list
        self._list.bind_model(metadata_list, _create_row)


def _create_row(metadata):
    return MetadataDetailsRow(key=metadata.key, value=metadata.value)



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Application window of Metadata Cleaner."""

from gettext import gettext as _
from typing import Any

from gi.repository import Adw, Gdk, Gio, GLib, GObject, Gtk, Pango
from metadatacleaner import RESOURCE_PATH
from metadatacleaner.modules.filestore import (
    SUPPORTED_FORMATS,
    FileStore,
    FileStoreState,
)
from metadatacleaner.modules.logger import Logger as logger
from metadatacleaner.ui.cleaningwarningdialog import CleaningWarningDialog
from metadatacleaner.ui.detailsview import DetailsView
from metadatacleaner.ui.dropoverlay import DropOverlay
from metadatacleaner.ui.filesview import FilesView
from metadatacleaner.ui.menubutton import MenuButton


@Gtk.Template(resource_path=RESOURCE_PATH + "/ui/Window.ui")
class Window(Adw.ApplicationWindow):
    """Application window of Metadata Cleaner."""

    __gtype_name__ = "Window"

    file_store = GObject.Property(type=FileStore, nick="file-store")

    _split_view: Adw.OverlaySplitView = Gtk.Template.Child()
    _view_stack: Gtk.Stack = Gtk.Template.Child()
    _details_view: DetailsView = Gtk.Template.Child()

    _file_chooser_dialog: Gtk.FileDialog = Gtk.Template.Child()
    _cleaning_warning_dialog: CleaningWarningDialog = Gtk.Template.Child()
    _header: Adw.HeaderBar = Gtk.Template.Child()
    _add_files_button: Gtk.MenuButton = Gtk.Template.Child()
    _drop_overlay: DropOverlay = Gtk.Template.Child()
    _toast_overlay: Adw.ToastOverlay = Gtk.Template.Child()
    _metadata_button: Gtk.ToggleButton = Gtk.Template.Child()
    _toolbar_view: Adw.ToolbarView = Gtk.Template.Child()

    def __init__(self, *args, **kwargs) -> None:
        """Window initialization."""
        app = kwargs["application"]
        super().__init__(title=app.name, *args, **kwargs)
        self._setup_size()
        self._setup_devel_style()
        self._setup_file_store()
        self._setup_drop_target()
        self._setup_actions()
        self._setup_file_filter()

    # SETUP #

    def _setup_file_filter(self) -> None:
        file_filter = Gtk.FileFilter()
        # translators: filter name while picking files
        file_filter.set_name(_("All supported files"))
        for mimetype, extensions in SUPPORTED_FORMATS.items():
            for extension in extensions:
                file_filter.add_suffix(extension[1:])
        self._file_chooser_dialog.set_default_filter(file_filter)

    def _setup_size(self) -> None:
        def on_close_request(window: Gtk.Window) -> None:
            width, height = self.get_default_size()
            maximized = self.is_maximized()
            self.get_application().settings.set_uint("window-width", width)
            self.get_application().settings.set_uint("window-height", height)
            self.get_application().settings.set_boolean(
                "window-maximized", maximized
            )

        self.set_default_size(
            self.get_application().settings.get_uint("window-width"),
            self.get_application().settings.get_uint("window-height"),
        )
        if self.get_application().settings.get_boolean("window-maximized"):
            self.maximize()
        self.connect("close-request", on_close_request)

    def _setup_devel_style(self) -> None:
        if self.get_application().devel:
            self.add_css_class("devel")

    def _setup_file_store(self) -> None:
        def on_items_changed(
            file_store: FileStore, position: int, added: int, removed: int
        ) -> None:
            if len(file_store) == 0:
                self.lookup_action("clean-metadata").set_enabled(False)
                self.close_details_view()
                self.show_empty_view()
            else:
                self.show_files_view()

        def on_state_changed(file_store: FileStore, file_index: int) -> None:
            self.lookup_action("clean-metadata").set_enabled(
                not (
                    file_store.state == FileStoreState.WORKING
                    or len(file_store.get_cleanable_files()) == 0
                )
            )

        self.file_store = FileStore()
        self.file_store.connect("items-changed", on_items_changed)
        self.file_store.connect("file-state-changed", on_state_changed)
        self.file_store.connect("state-changed", on_state_changed)
        self.get_application().settings.bind(
            "lightweight-cleaning",
            self.file_store,
            "lightweight-mode",
            Gio.SettingsBindFlags.DEFAULT,
        )

    def _setup_drop_target(self) -> None:
        def on_drop(widget: Gtk.DropTarget, value: Any, x: int, y: int):
            self._drop_overlay.dropping = False
            if isinstance(value, Gdk.FileList):
                self.file_store.add_gfiles(value.get_files())

        def on_leave(widget: Gtk.DropTarget):
            self._drop_overlay.dropping = False

        def on_enter(widget: Gtk.DropTarget, x: int, y: int):
            self._drop_overlay.dropping = True
            return Gdk.DragAction.COPY

        drop_target = Gtk.DropTarget.new(Gdk.FileList, Gdk.DragAction.COPY)
        drop_target.connect("drop", on_drop)
        drop_target.connect("enter", on_enter)
        drop_target.connect("leave", on_leave)
        self.add_controller(drop_target)

    def _present_about(self) -> None:
        app = self.get_application()
        about = Adw.AboutDialog.new()
        about.set_application_icon("io.gitlab.metadatacleaner.metadatacleaner")
        about.set_application_name(app.name)
        about.set_version(app.version)
        about.set_license_type(Gtk.License.GPL_3_0)
        about.set_developer_name("Metadata Cleaner Developers")
        about.set_website(
            "https://gitlab.com/metadatacleaner/metadatacleaner/"
        )
        about.set_issue_url(
            "https://gitlab.com/metadatacleaner/metadatacleaner/-/issues"
        )
        about.add_link(
            # translators: about dialog button that opens weblate
            _("Translate"),
            "https://hosted.weblate.org/engage/metadata-cleaner/",
        )
        about.set_support_url(
            "https://matrix.to/#/#metadata-cleaner:gnome.org"
        )
        about.set_debug_info(
            (
                "os: {os_name} {os_version}\n"
                "version: {version} ({profile})\n"
                "gtk: {gtk_rt_major}.{gtk_rt_minor}.{gtk_rt_micro} "
                "({gtk_ct_major}.{gtk_ct_minor}.{gtk_ct_micro})\n"
                "libadwaita: {adw_rt_major}.{adw_rt_minor}.{adw_rt_micro} "
                "({adw_ct_major}.{adw_ct_minor}.{adw_ct_micro})"
            ).format(
                os_name=GLib.get_os_info("NAME"),
                os_version=GLib.get_os_info("VERSION"),
                version=app.version,
                profile="devel" if app.devel else "prod",
                gtk_rt_major=Gtk.get_major_version(),
                gtk_rt_minor=Gtk.get_minor_version(),
                gtk_rt_micro=Gtk.get_micro_version(),
                gtk_ct_major=Gtk.MAJOR_VERSION,
                gtk_ct_minor=Gtk.MINOR_VERSION,
                gtk_ct_micro=Gtk.MICRO_VERSION,
                adw_rt_major=Adw.get_major_version(),
                adw_rt_minor=Adw.get_minor_version(),
                adw_rt_micro=Adw.get_micro_version(),
                adw_ct_major=Adw.MAJOR_VERSION,
                adw_ct_minor=Adw.MINOR_VERSION,
                adw_ct_micro=Adw.MICRO_VERSION,
            )
        )
        about.set_debug_info_filename("metadata-cleaner.txt")

        about.set_artists(
            [
                "Romain Vigier (former developer)",
                "GNOME Design Team https://gitlab.gnome.org/Teams/Design",
            ]
        )
        about.set_developers(
            [
                "Metadata Cleaner Developers",
                "segfault",
                'Evangelos "GeopJr" Paterakis https://geopjr.dev/',
                "Romain Vigier (former developer)",
            ]
        )

        # translators: Translate this string as your translator credits.
        # Name only:    Jane Doe
        # Name + URL:   Jane Doe https://example.com/
        # Name + Email: Jane Doe <janedoe@example.com>
        # Do not remove existing names.
        # Names are separated with newlines.
        about.set_translator_credits(_("translator-credits"))

        about.add_legal_section(
            title="mat2",
            copyright="Copyright 2018 Julien (jvoisin) Voisin julien.voisin+"
            "mat2@dustri.org\nCopyright 2016 Marie-Rose for mat2's logo",
            license_type=Gtk.License.LGPL_3_0,
        )
        about.present(self)

    def _setup_actions(self) -> None:
        def on_info(action: Gio.Action, parameters: None) -> None:
            self._on_show_info_dialog()

        info = Gio.SimpleAction.new("info", None)
        info.connect("activate", on_info)
        self.add_action(info)

        def on_close(action: Gio.Action, parameters: None) -> None:
            self.destroy()

        close = Gio.SimpleAction.new("close", None)
        close.connect("activate", on_close)
        self.add_action(close)

        def on_about(action: Gio.Action, parameters: None) -> None:
            self._present_about()

        about = Gio.SimpleAction.new("about", None)
        about.connect("activate", on_about)
        self.add_action(about)

        def on_toast(action: Gio.Action, parameters: GLib.Variant) -> None:
            self._toast_overlay.add_toast(
                Adw.Toast.new(title=parameters.get_string())
            )

        toast = Gio.SimpleAction.new("toast", GLib.VariantType.new("s"))
        toast.connect("activate", on_toast)
        self.add_action(toast)

        def on_add_files(action: Gio.Action, parameters: None) -> None:
            self._file_chooser_dialog.open_multiple(
                self, None, self._on_file_chooser_dialog_response
            )

        add_files = Gio.SimpleAction.new("add-files", None)
        add_files.connect("activate", on_add_files)
        self.add_action(add_files)

        def on_add_folders(action: Gio.Action, parameters: None) -> None:
            self._file_chooser_dialog.select_multiple_folders(
                self, None, self._on_folder_chooser_dialog_response
            )

        add_folders = Gio.SimpleAction.new("add-folders", None)
        add_folders.connect("activate", on_add_folders)
        self.add_action(add_folders)

        def on_remove_file(
            action: Gio.Action, parameters: GLib.Variant
        ) -> None:
            self.file_store.remove_file_with_index(parameters.get_uint32())

        remove_file = Gio.SimpleAction.new(
            "remove-file", GLib.VariantType.new("u")
        )
        remove_file.connect("activate", on_remove_file)
        self.add_action(remove_file)

        def on_clear_files(action: Gio.Action, parameters: None) -> None:
            self.file_store.remove_files()

        clear_files = Gio.SimpleAction.new("clear-files", None)
        clear_files.connect("activate", on_clear_files)
        clear_files.set_enabled(False)
        self.add_action(clear_files)

        def on_view_details(
            action: Gio.Action, parameters: GLib.Variant
        ) -> None:
            f = self.file_store.get_file_with_index(parameters.get_uint32())
            self._details_view.view_file(f)
            self.open_details_view()

        view_details = Gio.SimpleAction.new(
            "view-details", GLib.VariantType.new("u")
        )
        view_details.connect("activate", on_view_details)
        self.add_action(view_details)

        def on_toggle_details_view(
            action: Gio.Action, parameters: None
        ) -> None:
            self._split_view.set_show_sidebar(
                not self._split_view.get_show_sidebar()
            )

        toggle_details_view = Gio.SimpleAction.new("toggle-details-view", None)
        toggle_details_view.connect("activate", on_toggle_details_view)
        self.add_action(toggle_details_view)

        def on_clean_metadata(action: Gio.Action, parameters: None) -> None:
            self.close_details_view()
            if not self.get_application().settings.get_boolean(
                "cleaning-without-warning"
            ):
                self._cleaning_warning_dialog.choose(self)
                return
            self.file_store.clean_files()

        clean_metadata = Gio.SimpleAction.new("clean-metadata", None)
        clean_metadata.connect("activate", on_clean_metadata)
        clean_metadata.set_enabled(False)
        self.add_action(clean_metadata)

    # SIGNALS

    @Gtk.Template.Callback()
    def _on_sidebar_toggled(
        self, widget: Gtk.Widget, pspec: GObject.ParamSpec
    ) -> None:
        if self._split_view.get_show_sidebar() is False:
            self._view_stack.get_child_by_name("files").clear_selected_file()
            self._details_view.clear()

    def _on_file_chooser_dialog_response(
        self, file_dialog: Gtk.FileDialog, result: Gio.AsyncResult
    ) -> None:
        try:
            self.file_store.add_gfiles(
                file_dialog.open_multiple_finish(result)
            )
        except GLib.Error as e:
            if e.code == 2:
                # dismissed
                return

            logger.error(f"Couldn't open files: {e.code} {e.message}")
            return

    def _on_folder_chooser_dialog_response(
        self, file_dialog: Gtk.FileDialog, result: Gio.AsyncResult
    ) -> None:
        try:
            self.file_store.add_gfiles(
                file_dialog.select_multiple_folders_finish(result), True
            )
        except GLib.Error as e:
            if e.code == 2:
                # dismissed
                return

            logger.error(f"Couldn't open folders: {e.code} {e.message}")
            return

    @Gtk.Template.Callback()
    def _on_cleaning_warning_dialog_response(
        self, dialog: CleaningWarningDialog, response: str
    ) -> None:
        if response == "clean":
            self.file_store.clean_files()

    def _on_show_info_dialog(self) -> None:
        dlg = Adw.Dialog.new()
        # translators: dialog title that has some info about the
        #              cleaning process
        dlg.set_title(_("Info"))
        dlg.set_content_width(700)
        dlg.set_content_height(600)
        box = Gtk.Box.new(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        box.append(
            Gtk.Label(
                # translators: title for a section about the cleaning process
                label=_("Metadata & Privacy"),
                wrap=True,
                wrap_mode=Pango.WrapMode.WORD_CHAR,
                xalign=0,
                css_classes=["title-4"],
                margin_top=6,
            )
        )
        box.append(
            Gtk.Label(
                label=_(
                    # translators: paragraphs that describe the cleaning
                    #              process. Please keep the newlines as is
                    #              `\n`. If the language seems too advanced,
                    #              please simplify it while staying true to
                    #              the original; don't assume that the user
                    #              has any knowledge on what metadata is.
                    "Metadata consists of information that characterizes "
                    "data. Metadata is used to provide documentation for "
                    "data products. In essence, metadata answers who, what"
                    ", when, where, why, and how about every facet of the "
                    "data that is being documented.\n\nMetadata within a file "
                    "can tell a lot about you. Cameras record data about "
                    "when and where a picture was taken and which camera was"
                    " used. Office applications automatically add author and"
                    " company information to documents and spreadsheets. This"
                    " is sensitive information and you may not want to"
                    " disclose it."
                ),
                xalign=0,
                wrap=True,
                wrap_mode=Pango.WrapMode.WORD_CHAR,
            )
        )
        box.append(
            Gtk.Label(
                # translators: title for a section about the cleaning process
                label=_("Cleaning Process"),
                wrap=True,
                wrap_mode=Pango.WrapMode.WORD_CHAR,
                xalign=0,
                css_classes=["title-4"],
                margin_top=18,
            )
        )
        box.append(
            Gtk.Label(
                label=_(
                    # translators: paragraphs that describe the cleaning
                    #              process. Please keep the newlines as is
                    #              `\n`. If the language seems too advanced,
                    #              please simplify it while staying true to
                    #              the original; don't assume that the user
                    #              has any knowledge on what metadata is.
                    #              "Metadata Cleaner" is the app name. File
                    #              formats as in "pdf", "png", "music files"
                    #              etc. this section describes the case where
                    #              we might not know what a file actually is
                    #              and how to clean it.
                    "While Metadata Cleaner is doing its very best to display "
                    "metadata, it doesn't mean that a file is clean "
                    "from metadata if it doesn't show any. There is no "
                    "reliable way to detect every single possible metadata "
                    "for complex file formats. This is why you shouldn't rely"
                    " on metadata's presence to decide if your file must be"
                    " cleaned or not.\n\nMetadata Cleaner takes the content of"
                    "  the file and puts it into a new one without metadata,"
                    " ensuring that any undetected metadata is stripped."
                ),
                xalign=0,
                wrap=True,
                wrap_mode=Pango.WrapMode.WORD_CHAR,
            )
        )
        box.append(
            Gtk.Label(
                # translators: title for a section about the cleaning process
                label=_("Limitations"),
                wrap=True,
                wrap_mode=Pango.WrapMode.WORD_CHAR,
                xalign=0,
                css_classes=["title-4"],
                margin_top=18,
            )
        )
        box.append(
            Gtk.Label(
                label=_(
                    # translators: paragraphs that describe the cleaning
                    #              process. Please keep the newlines as
                    #              is `\n`. If the language seems too
                    #              advanced, please simplify it while
                    #              staying true to the original; don't
                    #              assume that the user has any knowledge
                    #              on what metadata is. "Metadata Cleaner"
                    #              is the app name.
                    "Be aware that metadata is not the only way of marking a"
                    " file. If the content itself discloses personal "
                    "information or has been watermarked, traditionally or "
                    "via steganography, Metadata Cleaner will not protect you."
                ),
                xalign=0,
                wrap=True,
                wrap_mode=Pango.WrapMode.WORD_CHAR,
            )
        )
        tbarview = Adw.ToolbarView(
            content=Gtk.ScrolledWindow(
                propagate_natural_height=True,
                propagate_natural_width=True,
                child=Adw.Clamp(
                    child=box,
                    margin_start=6,
                    margin_end=6,
                    margin_bottom=6,
                    tightening_threshold=200,
                ),
            )
        )
        tbarview.add_top_bar(Adw.HeaderBar.new())
        dlg.set_child(tbarview)
        dlg.present(self)

    # VIEWS #

    def show_empty_view(self) -> None:
        """Show an empty view."""
        self._add_files_button.set_visible(False)
        self._view_stack.set_visible_child_name("empty")
        self._header.set_show_title(False)
        self._metadata_button.set_visible(False)
        self._toolbar_view.set_top_bar_style(Adw.ToolbarStyle.FLAT)
        self.lookup_action("clear-files").set_enabled(False)

    def show_files_view(self) -> None:
        """Show the files."""
        self._add_files_button.set_visible(True)
        self._view_stack.set_visible_child_name("files")
        self._header.set_show_title(True)
        self._metadata_button.set_visible(True)
        self._toolbar_view.set_top_bar_style(Adw.ToolbarStyle.RAISED_BORDER)
        self.lookup_action("clear-files").set_enabled(True)

    def open_details_view(self) -> None:
        """Show the details view."""
        self._split_view.set_show_sidebar(True)

    def close_details_view(self) -> None:
        """Close the details view."""
        self._split_view.set_show_sidebar(False)



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Button to clean metadata of all added files."""

from gi.repository import Gtk
from metadatacleaner import RESOURCE_PATH


@Gtk.Template(resource_path=RESOURCE_PATH + "/ui/CleanMetadataButton.ui")
class CleanMetadataButton(Gtk.Button):
    """Button to clean metadata of all added files."""

    __gtype_name__ = "CleanMetadataButton"



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Row representing a file."""

from gi.repository import GLib, GObject, Gtk
from metadatacleaner import RESOURCE_PATH
from metadatacleaner.modules.file import File
from metadatacleaner.ui.badge import Badge


@Gtk.Template(resource_path=RESOURCE_PATH + "/ui/FileRow.ui")
class FileRow(Gtk.Box):
    """Row representing a file."""

    __gtype_name__ = "FileRow"

    file: File = GObject.Property(type=File)
    position: int = GObject.Property(type=int)

    @Gtk.Template.Callback()
    def _on_remove_file_button_clicked(self, button: Gtk.Button) -> None:
        self.activate_action(
            "win.remove-file", GLib.Variant.new_uint32(self.position)
        )



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Row displaying a metadata."""

from gi.repository import Adw, GObject, Gtk
from metadatacleaner import RESOURCE_PATH


@Gtk.Template(resource_path=RESOURCE_PATH + "/ui/MetadataDetailsRow.ui")
class MetadataDetailsRow(Adw.ActionRow):
    """Row displaying a metadata."""

    __gtype_name__ = "MetadataDetailsRow"

    key = GObject.Property(type=str, default="")
    value = GObject.Property(type=str, default="")



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Dialog warning the user of possible data loss on cleaning."""

from gi.repository import Adw, Gio, GObject, Gtk
from metadatacleaner import RESOURCE_PATH


@Gtk.Template(resource_path=RESOURCE_PATH + "/ui/CleaningWarningDialog.ui")
class CleaningWarningDialog(Adw.AlertDialog):
    """Dialog warning the user of possible data loss on cleaning."""

    __gtype_name__ = "CleaningWarningDialog"

    settings = GObject.Property(type=Gio.Settings)

    _checkbutton: Gtk.CheckButton = Gtk.Template.Child()

    @Gtk.Template.Callback()
    def _on_settings_changed(
        self, checkbox, p_spec: GObject.ParamSpec
    ) -> None:
        if not self.settings:
            return
        self.settings.set_boolean(
            "cleaning-without-warning", self._checkbutton.get_active()
        )



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Settings button."""

from gi.repository import Adw, Gtk
from metadatacleaner import RESOURCE_PATH


@Gtk.Template(resource_path=RESOURCE_PATH + "/ui/SettingsButton.ui")
class SettingsButton(Gtk.Button):
    """Settings button."""

    __gtype_name__ = "SettingsButton"

    _preferences_dlg: Adw.PreferencesDialog = Gtk.Template.Child()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.connect("clicked", self.show_pref_dialog)

    def show_pref_dialog(self, button: Gtk.Button) -> None:
        self._preferences_dlg.present(self.get_root())



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Logging facility using GLib."""

import inspect

from gi.repository import GLib, GObject


class Logger(GObject.GObject):
    """Logging facility using GLib."""

    _DOMAIN = "io.gitlab.metadatacleaner.metadatacleaner"

    @staticmethod
    def _log(message: str, level: GLib.LogLevelFlags) -> None:
        caller_frame = inspect.stack()[2]
        gvariant_dict = GLib.Variant(
            "a{sv}",
            {
                "MESSAGE": GLib.Variant("s", message),
                "CODE_FILE": GLib.Variant("s", caller_frame.filename),
                "CODE_LINE": GLib.Variant("i", caller_frame.lineno),
                "CODE_FUNC": GLib.Variant("s", caller_frame.function),
            },
        )
        GLib.log_variant(Logger._DOMAIN, level, gvariant_dict)

    @staticmethod
    def debug(message: str) -> None:
        """Print debug message.

        Args:
            message (str): Message to print.
        """
        Logger._log(message, GLib.LogLevelFlags.LEVEL_DEBUG)

    @staticmethod
    def warning(message: str) -> None:
        """Print warning message.

        Args:
            message (str): Message to print.
        """
        Logger._log(message, GLib.LogLevelFlags.LEVEL_WARNING)

    @staticmethod
    def message(message: str) -> None:
        """Print message.

        Args:
            message (str): Message to print.
        """
        Logger._log(message, GLib.LogLevelFlags.LEVEL_MESSAGE)

    @staticmethod
    def error(message: str) -> None:
        """Print error message.

        Args:
            message (str): Message to print.
        """
        Logger._log(message, GLib.LogLevelFlags.LEVEL_ERROR)

    @staticmethod
    def info(message: str) -> None:
        """Print informational message.

        Args:
            message (str): Message to print.
        """
        Logger._log(message, GLib.LogLevelFlags.LEVEL_INFO)

    @staticmethod
    def critical(message: str) -> None:
        """Print critical message.

        Args:
            message (str): Message to print.
        """
        Logger._log(message, GLib.LogLevelFlags.LEVEL_CRITICAL)



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Indicator showing the status of the files manager."""

from gettext import gettext as _
from gettext import ngettext

from gi.repository import Gio, GLib, GObject, Gtk
from metadatacleaner import RESOURCE_PATH
from metadatacleaner.modules.filestore import (
    FileStore,
    FileStoreAction,
    FileStoreState,
)
from metadatacleaner.ui.cleanmetadatabutton import CleanMetadataButton
from metadatacleaner.ui.settingsbutton import SettingsButton


@Gtk.Template(resource_path=RESOURCE_PATH + "/ui/StatusIndicator.ui")
class StatusIndicator(Gtk.Stack):
    """Indicator showing the status of the files manager."""

    __gtype_name__ = "StatusIndicator"

    file_store: FileStore = GObject.Property(type=FileStore, nick="file-store")

    _progressbar: Gtk.ProgressBar = Gtk.Template.Child()

    def __init__(self, *args, **kwargs) -> None:
        """Status indicator initialization."""
        super().__init__(*args, **kwargs)
        self.show_idle()

    def _sync_progressbar(self, current, total) -> None:
        if not self.file_store or not self.file_store.last_action:
            return
        text = {
            # translators: progress bar subtitle while loading files
            FileStoreAction.ADDING: _("Processing file {}/{}").format(
                current, total
            ),
            # translators: progress bar subtitle while cleaning files
            FileStoreAction.CLEANING: _("Cleaning file {}/{}").format(
                current, total
            ),
        }
        self._progressbar.set_text(text[self.file_store.last_action])
        self._progressbar.set_fraction(current / total if total > 0 else 0)

    @Gtk.Template.Callback()
    def _on_file_store_changed(
        self, widget: Gtk.Widget, pspec: GObject.ParamSpec
    ) -> None:
        self.file_store.connect(
            "state-changed", self._on_file_store_state_changed
        )
        self.file_store.connect(
            "progress-changed", self._on_file_store_progress_changed
        )

    @Gtk.Template.Callback()
    def _on_cancel_button_clicked(self, button: Gtk.Button) -> None:
        if self.file_store.last_action == FileStoreAction.ADDING:
            self.file_store.cancel_addding_gfiles()
        elif self.file_store.last_action == FileStoreAction.CLEANING:
            self.file_store.cancel_cleaning_files()

    def _on_file_store_state_changed(
        self, file_store: FileStore, new_state: FileStoreState
    ) -> None:
        if new_state == FileStoreState.WORKING:
            self.show_progressbar()

    def _on_file_store_progress_changed(
        self, file_store: FileStore, current: int, total: int
    ) -> None:
        self._sync_progressbar(current, total)
        if current == total:
            if (
                file_store.last_action == FileStoreAction.CLEANING
                and total != 0
                and self.get_visible_child_name() != "buttons"
            ):
                clean_message = ngettext(
                    "%i file cleaned",
                    "%i files cleaned",
                    len(file_store.get_cleaned_files()),
                ) % len(file_store.get_cleaned_files())
                error_message = (
                    ngettext(
                        "%i error occurred",
                        "%i errors occurred",
                        len(file_store.get_errored_files()),
                    )
                    % len(file_store.get_errored_files())
                    if len(file_store.get_errored_files()) > 0
                    else ""
                )
                body = " ".join([clean_message, error_message])
                self.send_toast(body)
                if not self.get_root().is_active():
                    self.send_done_notification(body)
            self.show_idle()

    def show_idle(self) -> None:
        """Show the idle state."""
        self.set_visible_child_name("buttons")

    def show_progressbar(self) -> None:
        """Show a progress bar."""
        self.set_visible_child_name("working")

    def send_toast(self, body: str) -> None:
        self.activate_action("win.toast", GLib.Variant.new_string(body))

    def send_done_notification(self, body: str) -> None:
        """Send a notification about the finished cleaning process."""
        window = self.get_root()
        app = window.get_application()
        notification = Gio.Notification.new(app.name)
        notification.set_body(body)
        notification.set_default_action_and_target(
            "app.show-window", GLib.Variant.new_uint32(window.get_id())
        )
        app.send_notification(f"done{window.get_id()}", notification)



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Files Manager object and states."""

import mimetypes
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor, as_completed
from enum import IntEnum, auto
from threading import Thread

import libmat2
from gi.repository import Gio, GLib, GObject
from metadatacleaner.modules.file import File, FileState
from metadatacleaner.modules.logger import Logger as logger


def _get_supported_formats() -> dict:
    formats = {}
    for parser in libmat2.parser_factory._get_parsers():
        for mimetype in parser.mimetypes:
            extensions = set()
            for extension in mimetypes.guess_all_extensions(mimetype):
                if extension not in libmat2.UNSUPPORTED_EXTENSIONS:
                    extensions.add(extension)
            if not extensions:
                continue
            formats[mimetype] = extensions
    return formats


SUPPORTED_FORMATS = _get_supported_formats()


class FileStoreState(IntEnum):
    """States the Files Manager can have."""

    IDLE = auto()
    WORKING = auto()


class FileStoreAction(IntEnum):
    """Actions the Files manager can do."""

    ADDING = auto()
    CLEANING = auto()


class FileStore(Gio.ListStore):
    """File Store object."""

    __gtype_name__ = "FileStore"

    __gsignals__ = {
        "file-state-changed": (GObject.SIGNAL_RUN_LAST, None, (int,)),
        "state-changed": (GObject.SIGNAL_RUN_LAST, None, (int,)),
        "progress-changed": (GObject.SIGNAL_RUN_LAST, None, (int, int)),
    }

    lightweight_mode: bool = GObject.Property(
        type=bool, nick="lightweight-mode", default=False
    )

    def __init__(self) -> None:
        """File Store initialization."""
        Gio.ListStore.__init__(self, item_type=File)
        self.state = FileStoreState.IDLE
        self.last_action: FileStoreAction | None = None
        self.progress = (0, 0)
        self.add_files_executor = ThreadPoolExecutor()
        self.clean_files_executor = ThreadPoolExecutor()

    def _on_file_state_changed(self, f: File, new_state: FileState) -> None:
        def emit() -> bool:
            self.emit("file-state-changed", self.get_index_of_file(f))
            return GLib.SOURCE_REMOVE

        GLib.idle_add(emit)

    def _set_state(self, state: FileStoreState) -> None:
        if state == self.state:
            return
        self.state = state

        def emit() -> bool:
            self.emit("state-changed", state)
            return GLib.SOURCE_REMOVE

        GLib.idle_add(emit)

    def _set_progress(self, current: int, total: int) -> None:
        self.progress = (current, total)

        def emit() -> bool:
            self.emit("progress-changed", current, total)
            return GLib.SOURCE_REMOVE

        GLib.idle_add(emit)

    def get_files(self) -> list[File]:
        """Get all the files from the File Store.

        Returns:
            List[File]: List of files.
        """
        return self

    def get_file_with_index(self, index: int) -> File:
        """Get a file at a specified index.

        Args:
            index (int): The desired index.

        Returns:
            Optional[File]: The file if found, else None.
        """
        f = self.get_item(index)
        if not f:
            raise RuntimeError("No file in the file store for this index.")
        return f

    def get_index_of_file(self, f: File) -> int:
        """Get the index of a given file.

        Args:
            f (File): The file to get the index for.

        Returns:
            int: The file index.
        """
        found, position = self.find(f)
        if not found:
            raise RuntimeError("File not found in file store.")
        return position

    def add_gfiles(
        self, gfiles: list[Gio.File], recursive: bool = True
    ) -> None:
        """Add Gio Files to the Files Manager.

        Args:
            gfiles (List[Gio.File]): List of Gio Files to add.
            recursive (bool, optional): If subdirectories should also be looked
            into. Defaults to True.
        """
        thread = Thread(
            target=self._add_gfiles_async,
            args=(gfiles, recursive),
            daemon=True,
        )
        thread.start()

    def _add_gfiles_async(
        self, gfiles: list[Gio.File], recursive: bool = True
    ) -> None:
        self._set_state(FileStoreState.WORKING)
        all_gfiles = self._gather_all_gfiles(gfiles, recursive)
        self._set_progress(
            self.progress[0], self.progress[1] + len(all_gfiles)
        )
        self.last_action = FileStoreAction.ADDING
        futures = {
            self.add_files_executor.submit(self._add_gfile, gfile)
            for gfile in all_gfiles
        }
        for future in as_completed(futures):
            current = self.progress[0] + 1
            total = self.progress[1]
            self._set_progress(current, total)
            if current == total:
                self._stop_adding_gfiles()

    def _gather_all_gfiles(
        self, gfiles: list[Gio.File], recursive: bool = True
    ) -> list[Gio.File]:
        all_gfiles: list[Gio.File] = []
        for gfile in gfiles:
            if not gfile:
                continue
            f_type = gfile.query_file_type(Gio.FileQueryInfoFlags.NONE, None)
            if f_type == Gio.FileType.DIRECTORY:
                all_gfiles.extend(self._get_gfiles_from_dir(gfile, recursive))
            elif f_type == Gio.FileType.REGULAR:
                all_gfiles.append(gfile)
            else:
                logger.warning(
                    f"File {gfile.get_path()} is neither a directory nor a "
                    "regular file, skipping."
                )
        return all_gfiles

    def _get_gfiles_from_dir(
        self, dir: Gio.File, recursive: bool
    ) -> list[Gio.File]:
        gfiles: list[Gio.File] = []
        subdirs: list[Gio.File] = []
        children_enumerator = dir.enumerate_children(
            "", Gio.FileQueryInfoFlags.NONE, None
        )
        while True:
            info = children_enumerator.next_file(None)
            if info is None:
                break
            child = children_enumerator.get_child(info)
            if info.get_file_type() == Gio.FileType.DIRECTORY:
                if recursive:
                    subdirs.append(child)
            elif info.get_file_type() == Gio.FileType.REGULAR:
                gfiles.append(child)
        children_enumerator.close(None)
        with ThreadPoolExecutor() as executor:
            for subgfiles in executor.map(
                self._get_gfiles_from_dir, subdirs, [recursive] * len(subdirs)
            ):
                gfiles.extend(subgfiles)
        return gfiles

    def _add_gfile(self, gfile: Gio.File) -> None:
        if gfile is None:
            logger.warning("File is None, skipping.")
            return

        if not gfile.query_exists(None):
            logger.warning(
                f"File {gfile.get_path()} does not exist, skipping."
            )
            return

        if bool(list(filter(lambda x: x.path == gfile.get_path(), self))):
            logger.warning(f"Skipping {gfile.get_path()}, already added.")
            return

        f = File(gfile)
        f.check_metadata()

        def finish() -> bool:
            self.append(f)
            f.connect("state-changed", self._on_file_state_changed)
            return GLib.SOURCE_REMOVE

        GLib.idle_add(finish)

    def _stop_adding_gfiles(self) -> None:
        self.add_files_executor.shutdown(wait=False, cancel_futures=True)
        self.add_files_executor = ThreadPoolExecutor()
        self._set_state(FileStoreState.IDLE)
        self._set_progress(0, 0)

    def cancel_addding_gfiles(self) -> None:
        """Cancel adding GFiles."""
        self._stop_adding_gfiles()

    def remove_file(self, f: File) -> None:
        """Remove a file from the File Store.

        Args:
            f (File): The file to remove.
        """
        self.remove_file_with_index(self.get_index_of_file(f))

    def remove_file_with_index(self, index: int) -> None:
        """Remove the file with the given index from the File Store.

        Args:
            index (int): The index of the file to remove.
        """
        self.remove(index)

    def remove_files(self) -> None:
        """Remove all the files from the File Store."""
        self.remove_all()

    def clean_files(self) -> None:
        """Remove metadata from all the cleanable files."""
        thread = Thread(target=self._clean_files_async, daemon=True)
        thread.start()

    def _clean_files_async(self) -> None:
        cleanable_files = self.get_cleanable_files()
        self._set_progress(
            self.progress[0], self.progress[1] + len(cleanable_files)
        )
        self._set_state(FileStoreState.WORKING)
        self.last_action = FileStoreAction.CLEANING
        futures = {
            self.clean_files_executor.submit(f.clean, self.lightweight_mode)
            for f in cleanable_files
        }
        for future in as_completed(futures):
            current = self.progress[0] + 1
            total = self.progress[1]
            self._set_progress(current, total)
        self._stop_cleaning_files()

    def _stop_cleaning_files(self) -> None:
        self.clean_files_executor.shutdown(wait=False, cancel_futures=True)
        self.clean_files_executor = ThreadPoolExecutor()
        self._set_state(FileStoreState.IDLE)
        self._set_progress(0, 0)

    def cancel_cleaning_files(self) -> None:
        """Cancel the cleaning process."""
        self._stop_cleaning_files()

    def get_cleanable_files(self) -> list[File]:
        """Get all the cleanable files.

        Returns:
            List[File]: List of cleanable files.
        """
        return self._get_files_with_states(
            (FileState.HAS_METADATA, FileState.HAS_NO_METADATA)
        )

    def get_cleaned_files(self) -> list[File]:
        """Get all the cleaned files.

        Returns:
            List[File]: List of cleaned files.
        """
        return self._get_files_with_states((FileState.CLEANED,))

    def get_errored_files(self) -> list[File]:
        """Get all the files with errors.

        Returns:
            List[File]: List of files with errors.
        """
        return self._get_files_with_states(
            (
                FileState.ERROR_WHILE_INITIALIZING,
                FileState.ERROR_WHILE_CHECKING_METADATA,
                FileState.ERROR_WHILE_REMOVING_METADATA,
            )
        )

    def _get_files_with_states(
        self, states: Iterable[FileState]
    ) -> list[File]:
        return list(filter(lambda f: f.state in states, self))



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""View showing all the files."""

from gi.repository import Adw, Gio, GLib, GObject, Gtk
from metadatacleaner import RESOURCE_PATH
from metadatacleaner.modules.filestore import FileStore
from metadatacleaner.ui.filerow import FileRow
from metadatacleaner.ui.statusindicator import StatusIndicator


@Gtk.Template(resource_path=RESOURCE_PATH + "/ui/FilesView.ui")
class FilesView(Adw.Bin):
    """View showing all the files."""

    __gtype_name__ = "FilesView"

    file_store = GObject.Property(type=FileStore, nick="file-store")

    _selection_model: Gtk.SingleSelection = Gtk.Template.Child()

    def __init__(self, *args, **kwargs) -> None:
        """View initialization."""
        super().__init__(*args, **kwargs)
        self._setup_actions()

    def _setup_actions(self) -> None:
        group = Gio.SimpleActionGroup.new()
        self.insert_action_group("files", group)

        def on_select(action: Gio.Action, parameters: GLib.Variant) -> None:
            self._selection_model.set_selected(parameters.get_uint32())

        select = Gio.SimpleAction.new("select", GLib.VariantType.new("u"))
        select.connect("activate", on_select)
        group.insert(select)

    @Gtk.Template.Callback()
    def _on_selection_changed(
        self, selection_model: Gtk.SelectionModel, position: int, n_items: int
    ) -> None:
        number_of_items = len(selection_model.get_model())
        if selection_model.get_selected() >= number_of_items:
            return
        self.activate_action(
            "win.view-details",
            GLib.Variant.new_uint32(selection_model.get_selected()),
        )

    def get_selected_file_index(self) -> int:
        """Get the index of the currently selected file.

        Returns:
            int: Index of the selected file.
        """
        return self._selection_model.get_selected()

    def clear_selected_file(self) -> None:
        """Unselect the currently selected file."""
        number_of_items = len(self._selection_model.get_model())
        if self._selection_model.get_selected() >= number_of_items:
            return
        self._selection_model.unselect_item(
            self._selection_model.get_selected()
        )



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Menu button."""

from gi.repository import Gtk
from metadatacleaner import RESOURCE_PATH


@Gtk.Template(resource_path=RESOURCE_PATH + "/ui/MenuButton.ui")
class MenuButton(Gtk.MenuButton):
    """Menu button."""

    __gtype_name__ = "MenuButton"



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, GObject, Gtk


class DropOverlay(Adw.Bin):
    __gtype_name__ = "DropOverlay"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.overlay = Gtk.Overlay()
        self.status_page = Adw.StatusPage(
            icon_name="metadatacleaner-paper-symbolic",
            css_classes=["status"],
        )

        self.revealer = Gtk.Revealer(
            child=self.status_page,
            can_target=False,
            transition_type=Gtk.RevealerTransitionType.CROSSFADE,
            transition_duration=800,
        )

        self.overlay.add_overlay(self.revealer)
        self.set_child(self.overlay)
        self.add_css_class("dropoverlay")

    @GObject.Property(type=Gtk.Widget)
    def overlay_child(self) -> Gtk.Widget:
        return self.overlay.get_child()

    @overlay_child.setter  # type: ignore
    def overlay_child(self, value: Gtk.Widget):
        self.overlay.set_child(value)

    @GObject.Property(type=bool, default=False)
    def dropping(self) -> bool:
        return self.revealer.get_reveal_child()

    @dropping.setter  # type: ignore
    def dropping(self, value: bool):
        self.revealer.set_reveal_child(value)

    @GObject.Property(type=str)
    def title(self) -> str:
        return self.status_page.get_title()

    @title.setter  # type: ignore
    def title(self, value: str):
        self.status_page.set_title(value)

    @GObject.Property(type=bool, default=False)
    def compact(self) -> bool:
        return self.status_page.has_css_class("compact")

    @compact.setter  # type: ignore
    def compact(self, value: bool):
        if self.compact != value:
            if value:
                self.status_page.add_css_class("compact")
            else:
                self.status_page.remove_css_class("compact")

    @GObject.Property(type=str)
    def icon_name(self) -> str:
        return self.status_page.get_icon_name()

    @icon_name.setter  # type: ignore
    def icon_name(self, value: str):
        self.status_page.set_icon_name(value)



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""File object and states."""

import hashlib
import re
import tempfile
from enum import IntEnum, auto
from gettext import gettext as _
from pathlib import Path

from gi.repository import Gio, GLib, GObject
from libmat2 import parser_factory
from metadatacleaner.modules.logger import Logger as logger
from metadatacleaner.modules.metadata import (
    Metadata,
    MetadataFile,
    MetadataList,
    MetadataStore,
)


class FileState(IntEnum):
    """States that a File can have."""

    INITIALIZING = auto()
    ERROR_WHILE_INITIALIZING = auto()
    UNSUPPORTED = auto()
    SUPPORTED = auto()
    CHECKING_METADATA = auto()
    ERROR_WHILE_CHECKING_METADATA = auto()
    HAS_NO_METADATA = auto()
    HAS_METADATA = auto()
    REMOVING_METADATA = auto()
    ERROR_WHILE_REMOVING_METADATA = auto()
    CLEANED = auto()


class File(GObject.GObject):
    """File object."""

    __gtype_name__ = "File"

    __gsignals__ = {"state-changed": (GObject.SIGNAL_RUN_LAST, None, (int,))}

    filename = GObject.Property(type=str)
    directory = GObject.Property(type=str)
    display_directory = GObject.Property(
        type=bool, nick="display-directory", default=True
    )
    icon_name = GObject.Property(type=str, nick="icon-name")
    simple_state = GObject.Property(
        type=str, nick="simple-state", default="working"
    )
    metadata = GObject.Property(type=MetadataStore)
    total_metadata = GObject.Property(
        type=int, nick="total-metadata", default=0
    )
    selectable = GObject.Property(type=bool, default=False)
    message_type = GObject.Property(
        type=str, nick="message-type", default="none"
    )
    has_message = GObject.Property(
        type=bool, nick="has-message", default=False
    )

    def __init__(self, gfile: Gio.File) -> None:
        """File initialization.

        Args:
            gfile (Gio.File): The Gio File that the File will be built from.
        """
        super().__init__()
        self._gfile = gfile
        self._temp_path = self._compute_temp_path(gfile.get_path())
        self.path = gfile.get_path()
        self.filename = gfile.get_basename()
        self._extract_path(gfile)
        self.state = FileState.INITIALIZING
        self.mimetype = "text/plain"
        self.icon_name = Gio.content_type_get_generic_icon_name(self.mimetype)
        self.metadata = MetadataStore()
        self.error: Exception | None = None

    def _compute_temp_path(self, path: str) -> str:
        # We have to keep the extension so that ffmpeg doesn't break
        digest = hashlib.sha256(path.encode("utf-8")).hexdigest()
        return str(
            Path(tempfile.gettempdir()) / f"{digest}{Path(path).suffix}"
        )

    def _simplify_dir_path(self, path: str) -> None:
        dir_path = str(Path(path).parent)
        doc_path_match = re.match(r"/run/user/\d+/doc/[a-z\d]+/?", dir_path)
        if doc_path_match:
            # Remove the Document Store path
            dir_path = dir_path.replace(doc_path_match.group(0), "", 1)
        elif re.match(GLib.get_home_dir(), dir_path):
            # Replace the home path with the friendly ~
            dir_path = dir_path.replace(GLib.get_home_dir(), "~", 1)
        self.directory = dir_path
        self.display_directory = bool(dir_path)

    def _extract_path(self, gfile: Gio.File) -> None:
        self._simplify_dir_path(gfile.get_path())
        if gfile:
            gfile.query_info_async(
                "xattr::document-portal.host-path",
                Gio.FileQueryInfoFlags.NONE,
                GLib.PRIORITY_DEFAULT,
                None,
                self.on_gfile_info_cb,
            )

    def on_gfile_info_cb(
        self, gfile: Gio.File, result: Gio.AsyncResult
    ) -> None:
        try:
            info = gfile.query_info_finish(result)
            path = info.get_attribute_string(
                "xattr::document-portal.host-path"
            )
            if path:
                self._simplify_dir_path(path)
        except GLib.Error:
            logger.error("Could not query file: {e.code} {e.message}")

    def _set_state(self, state: FileState) -> None:
        if state == self.state or self.state == FileState.CLEANED:
            return

        def update_state(state) -> bool:
            simple_states = {
                FileState.INITIALIZING: "working",
                FileState.ERROR_WHILE_INITIALIZING: "error",
                FileState.UNSUPPORTED: "error",
                FileState.SUPPORTED: "working",
                FileState.CHECKING_METADATA: "working",
                FileState.ERROR_WHILE_CHECKING_METADATA: "error",
                FileState.HAS_NO_METADATA: "warning",
                FileState.HAS_METADATA: "has-metadata",
                FileState.REMOVING_METADATA: "working",
                FileState.ERROR_WHILE_REMOVING_METADATA: "error",
                FileState.CLEANED: "clean",
            }
            message_types = {
                FileState.INITIALIZING: "none",
                FileState.ERROR_WHILE_INITIALIZING: "error-initializing",
                FileState.UNSUPPORTED: "unsupported",
                FileState.SUPPORTED: "none",
                FileState.CHECKING_METADATA: "none",
                FileState.ERROR_WHILE_CHECKING_METADATA: "error-checking",
                FileState.HAS_NO_METADATA: "no-metadata",
                FileState.HAS_METADATA: "none",
                FileState.REMOVING_METADATA: "none",
                FileState.ERROR_WHILE_REMOVING_METADATA: "error-removing",
                FileState.CLEANED: "none",
            }
            self.simple_state = simple_states[state]
            self.selectable = state == FileState.HAS_METADATA
            self.message_type = message_types[state]
            self.has_message = self.message_type != "none"
            self.emit("state-changed", state)
            return GLib.SOURCE_REMOVE

        GLib.idle_add(update_state, state)
        self.state = state

    def check_metadata(self) -> None:
        """Set up the parser and check the metadata present in the file."""
        try:
            parser, mimetype = parser_factory.get_parser(self.path)
        except Exception as e:
            self._setup_parser_error(e)
        else:
            self._setup_parser_finish(parser, mimetype)

        if self.state != FileState.SUPPORTED:
            return
        self._set_state(FileState.CHECKING_METADATA)
        try:
            metadata = self._parser.get_meta()
        except Exception as e:
            self._check_metadata_error(e)
        else:
            self._check_metadata_finish(metadata)

    def _setup_parser_error(self, error: Exception) -> None:
        self.error = error
        logger.warning(
            f"Error while setting up parser for {self.filename}: {error}"
        )
        self._set_state(FileState.ERROR_WHILE_INITIALIZING)

    def _setup_parser_finish(self, parser, mimetype) -> None:
        self._parser = parser
        if mimetype:

            def update_mimetype(mimetype) -> bool:
                self.mimetype = mimetype
                self.icon_name = Gio.content_type_get_generic_icon_name(
                    self.mimetype
                )
                return GLib.SOURCE_REMOVE

            GLib.idle_add(update_mimetype, mimetype)
        if self._parser:
            self._set_state(FileState.SUPPORTED)
        else:
            self._set_state(FileState.UNSUPPORTED)

    def _check_metadata_error(self, error: Exception) -> None:
        self.error = error
        logger.warning(
            f"Error while checking metadata for {self.filename}: {error}"
        )
        self._set_state(FileState.ERROR_WHILE_CHECKING_METADATA)

    def _check_metadata_finish(self, metadata) -> None:
        if not bool(metadata):
            self._set_state(FileState.HAS_NO_METADATA)
            return
        total_metadata = 0
        # Metadata found in multiple files (e.g. in archive)
        if isinstance(metadata[list(metadata)[0]], dict):
            for filename, file_metadata in metadata.items():
                metadata_list = MetadataList()
                for key, value in file_metadata.items():
                    metadata_list.append(Metadata(key=key, value=value))
                self.metadata.append(
                    MetadataFile(
                        filename=Path(self.filename) / filename,
                        metadata=metadata_list,
                    )
                )
                total_metadata += len(metadata_list)
        # Metadata found in a single file
        else:
            metadata_list = MetadataList()
            for key, value in metadata.items():
                metadata_list.append(Metadata(key=key, value=value))
            self.metadata.append(
                MetadataFile(filename=self.filename, metadata=metadata_list)
            )
            total_metadata += len(metadata_list)

        def update_total_metadata(total_metadata) -> bool:
            self.total_metadata = total_metadata
            return GLib.SOURCE_REMOVE

        GLib.idle_add(update_total_metadata, total_metadata)
        self._set_state(FileState.HAS_METADATA)

    def clean(self, lightweight_mode=False) -> None:
        """Clean the metadata from the file.

        Args:
            lightweight_mode (bool, optional): Use mat2 lightweight mode to
                preserve data integrity. Defaults to False.
        """
        if self.state not in [
            FileState.HAS_METADATA,
            FileState.HAS_NO_METADATA,
        ]:
            return
        self._set_state(FileState.REMOVING_METADATA)
        try:
            self._parser.output_filename = self._temp_path
            self._parser.lightweight_cleaning = lightweight_mode
            result = self._parser.remove_all()
            if result is False:
                # translators: generic error message shown when the cleaning
                #              was unsuccessful but we don't know why
                raise RuntimeError(_("An error occurred during the cleaning."))
            if not Path(self._temp_path).exists():
                raise RuntimeError(
                    _(
                        # translators: error message when a file is no longer
                        #              available during cleaning. "Something
                        #              bad happened" as in "Something went
                        #              wrong"
                        "Something bad happened during the cleaning, "
                        "cleaned file not found"
                    )
                )
            cleaned_gfile = Gio.File.new_for_path(self._temp_path)
            cleaned_gfile.move(
                self._gfile, Gio.FileCopyFlags.OVERWRITE, None, None, None
            )
        except Exception as e:
            self._clean_error(e)
        else:
            self._clean_finish()

    def _clean_error(self, error: Exception) -> None:
        self.error = error
        logger.warning(
            f"Error while cleaning metadata from {self.filename}: {error}"
        )
        self._set_state(FileState.ERROR_WHILE_REMOVING_METADATA)

    def _clean_finish(self) -> None:
        self._set_state(FileState.CLEANED)



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Badge to display information."""

from gettext import gettext as _

from gi.repository import Adw, GObject, Gtk
from metadatacleaner import RESOURCE_PATH


@Gtk.Template(resource_path=RESOURCE_PATH + "/ui/Badge.ui")
class Badge(Adw.Bin):
    """Badge to display information."""

    __gtype_name__ = "Badge"
    _state = ""
    _label = ""

    @GObject.Property(type=str, default="")
    def label(self):
        """Get the badge label.

        Returns:
            str: The badge label.
        """
        return self._label

    @label.setter  # type: ignore
    def label(self, value):
        self._label = value

    @GObject.Property(type=str, default="")
    def state(self):
        """Get the current badge state.

        working => spinner
        warning => !
        error => -
        clean => checkmark
        has-metadata => label

        Returns:
            str: The badge state.
        """
        return self._state

    @state.setter  # type: ignore
    def state(self, value):
        if self._state == value:
            return
        self._state = value
        self._reset()
        match value:
            case "working":
                self.set_child(Gtk.Spinner(spinning=True))
            case "warning":
                self.add_css_class("warning")
                self.set_child(
                    Gtk.Image(
                        icon_name="metadatacleaner-warning-symbolic",
                        # translators: tooltip text on the badge that shows an
                        #              icon representing the state of the file
                        tooltip_text=_("Warning"),
                    )
                )
            case "error":
                self.add_css_class("error")
                self.set_child(
                    Gtk.Image(
                        icon_name="metadatacleaner-error-symbolic",
                        # translators: tooltip text on the badge that shows an
                        #              icon representing the state of the file
                        tooltip_text=_("Error"),
                    )
                )
            case "has-metadata":
                self.add_css_class("metadata")
                label = Gtk.Label(css_classes=["numeric"], label=self._label)
                self.bind_property("label", label, "label")
                self.set_child(label)
            case "clean":
                self.add_css_class("success")
                self.set_child(
                    Gtk.Image(
                        icon_name="metadatacleaner-ok-symbolic",
                        # translators: tooltip text on the badge that shows an
                        #              icon representing the state of the file
                        tooltip_text=_("Cleaned"),
                    )
                )

    def _reset(self):
        self.set_css_classes(["badge"])
        self.set_child(None)



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""View details of a file."""

from gettext import gettext as _
from typing import Optional

from gi.repository import Adw, GLib, Gtk
from metadatacleaner import RESOURCE_PATH
from metadatacleaner.modules.file import File, FileState
from metadatacleaner.ui.metadataview import MetadataView


@Gtk.Template(resource_path=RESOURCE_PATH + "/ui/DetailsView.ui")
class DetailsView(Adw.Bin):
    """View details of a file."""

    __gtype_name__ = "DetailsView"

    def __init__(self, *args, **kwargs) -> None:
        """Details view initialization."""
        super().__init__(*args, **kwargs)
        self.clear()

    def view_file(self, f: File) -> None:
        """Set the file to view.

        Args:
            f (File): The file to view.
        """
        self.set_child(None)
        if f.state == FileState.HAS_METADATA:
            self._setup_metadata_details(f)
        elif f.state == FileState.CLEANED:
            self._setup_cleaned_details()
        elif f.state in [
            FileState.ERROR_WHILE_CHECKING_METADATA,
            FileState.ERROR_WHILE_INITIALIZING,
            FileState.ERROR_WHILE_REMOVING_METADATA,
            FileState.HAS_NO_METADATA,
            FileState.UNSUPPORTED,
        ]:
            self._setup_error_details(f)

    def set_content(self, w) -> None:
        """Set the widget's child."""
        self.set_child(w)

    def _setup_cleaned_details(self) -> None:
        self.set_content(
            Adw.StatusPage(
                # translators: side panel message when selecting a file that
                #              has been cleaned
                title=_("The File Has Been Cleaned"),
                description=_(
                    # translators: side panel subtitle when selecting a file
                    #              that has been cleaned. The limitations can
                    #              be viewed by clicking a button below it
                    "Known metadata have been removed, however the cleaning "
                    "process has some limitations."
                ),
                child=Gtk.Button(
                    # translators: button that shows more info about the
                    #              cleaning process, it's shown below a
                    #              label that gives it context
                    label=_("Learn more"),
                    action_name="win.info",
                    halign=Gtk.Align.CENTER,
                    css_classes=["pill", "suggested-action"],
                ),
                css_classes=["compact"],
            )
        )

    def _setup_metadata_details(self, f: File) -> None:
        self.set_content(MetadataView(metadata=f.metadata))

    def _setup_error_details(self, f: File) -> None:
        info_titles = {
            # translators: error shown on the side panel when selecting
            #              a file
            FileState.ERROR_WHILE_INITIALIZING: _("Unable to Read the File"),
            # translators: error shown on the side panel when selecting
            #              a file
            FileState.UNSUPPORTED: _("File Type not Supported"),
            FileState.ERROR_WHILE_CHECKING_METADATA: _(
                # translators: error shown on the side panel when selecting
                #              a file
                "Unable to Check for Metadata"
            ),
            # translators: error shown on the side panel when selecting
            #              a file
            FileState.HAS_NO_METADATA: _("No Known Metadata"),
            FileState.ERROR_WHILE_REMOVING_METADATA: _(
                # translators: error shown on the side panel when
                #              selecting a file
                "Unable to Remove Metadata"
            ),
        }
        info_details = str(f.error or "")
        if f.state == FileState.HAS_NO_METADATA:
            # translators: message shown on the side panel when selecting a
            #              file that doesn't have known metadata
            info_details = _("The file will be cleaned anyway to be sure.")
        self.set_content(
            Adw.StatusPage(
                title=info_titles[f.state],
                description=info_details,
                css_classes=["compact"],
            )
        )

    def clear(self) -> None:
        """Informs the user that they need to select a file"""
        self.set_content(
            Adw.StatusPage(
                # translators: message shown on the side panel when no file
                #              has been selected
                title=_("Select a file to view its metadata"),
                css_classes=["compact"],
            )
        )



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""List of multiple files' metadata."""

from gi.repository import GObject, Gtk
from metadatacleaner import RESOURCE_PATH
from metadatacleaner.modules.metadata import MetadataStore
from metadatacleaner.ui.metadatadetails import MetadataDetails


@Gtk.Template(resource_path=RESOURCE_PATH + "/ui/MetadataView.ui")
class MetadataView(Gtk.ScrolledWindow):
    """List of multiple files' metadata."""

    __gtype_name__ = "MetadataView"

    metadata = GObject.Property(type=MetadataStore)



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Application for Metadata Cleaner."""

from gettext import gettext as _

from gi.repository import Adw, Gdk, Gio, GLib, GObject, Gtk
from metadatacleaner.ui.window import Window


class MetadataCleaner(Adw.Application):
    """Application for Metadata Cleaner."""

    __gtype_name__ = "MetadataCleaner"

    devel = GObject.Property(type=bool, default=False)
    settings = GObject.Property(type=Gio.Settings)
    version = GObject.Property(type=str)

    def __init__(self, devel: bool, version: str, *args, **kwargs) -> None:
        """Application initialization.

        Args:
            app_id (str): Application ID
            version (str): Version string
        """
        super().__init__(
            flags=Gio.ApplicationFlags.HANDLES_OPEN, *args, **kwargs
        )
        # translators: app name
        self.name = _("Metadata Cleaner")
        self.devel = devel
        self.version = version
        self.settings = Gio.Settings.new(self.get_application_id())

    # APPLICATION METHODS #

    def do_activate(self) -> None:
        """Run on application activation."""
        self.new_window()

    def do_startup(self) -> None:
        """Run on application startup."""
        Adw.Application.do_startup(self)
        self._setup_actions()
        self._setup_accels()

    def do_open(self, gfiles: list[Gio.File], n_files: int, hint: str) -> None:
        """Run when files are passed to the command line."""
        self.new_window(gfiles=gfiles)

    # SETUP #

    def _setup_actions(self) -> None:
        def on_show_window(
            action: Gio.Action, parameters: GLib.Variant
        ) -> None:
            window_id = parameters.get_uint32()
            window = self.get_window_by_id(window_id)
            window.present_with_time(Gdk.CURRENT_TIME)
            self.withdraw_notification(f"done{window_id}")

        show_window = Gio.SimpleAction.new(
            "show-window", GLib.VariantType.new("u")
        )
        show_window.connect("activate", on_show_window)
        self.add_action(show_window)

        def on_new_window(action: Gio.Action, parameters: None) -> None:
            self.new_window()

        new_window = Gio.SimpleAction.new("new-window", None)
        new_window.connect("activate", on_new_window)
        self.add_action(new_window)

        def on_quit_app(action: Gio.Action, parameters: None) -> None:
            for window in self.get_windows():
                self.withdraw_notification(f"done{window.get_id()}")
            self.quit()

        quit_app = Gio.SimpleAction.new("quit", None)
        quit_app.connect("activate", on_quit_app)
        self.add_action(quit_app)

        self.add_action(self.settings.create_action("lightweight-cleaning"))

    def _setup_accels(self) -> None:
        self.set_accels_for_action("app.new-window", ["<Control>n"])
        self.set_accels_for_action("app.quit", ["<Control>q"])
        self.set_accels_for_action("win.add-files", ["<Control>o"])
        self.set_accels_for_action("win.add-folders", ["<Control><Shift>o"])
        self.set_accels_for_action("win.clear-files", ["<Control>r"])
        self.set_accels_for_action("win.clean-metadata", ["<Control>m"])
        self.set_accels_for_action("win.close", ["<Control>w"])

    # PUBLIC #

    def new_window(self, gfiles: list[Gio.File] | None = None) -> None:
        """Create a new window.

        Args:
            gfiles (List[Gio.File], optional): List of files to be added to the
                new window. Defaults to None.
        """

        def on_window_destroyed(window) -> None:
            self.withdraw_notification(f"done{window.get_id()}")

        window = Window(application=self)
        window.connect("destroy", on_window_destroyed)
        window.present()
        if gfiles:
            window.file_store.add_gfiles(gfiles)



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Metadata classes."""

from gi.repository import Gio, GObject


class Metadata(GObject.GObject):
    """Metadata object."""

    __gtype_name__ = "Metadata"

    key = GObject.Property(type=str)
    value = GObject.Property(type=str)


class MetadataList(Gio.ListStore):
    """Metadata List object."""

    __gtype_name__ = "MetadataList"

    def __init__(self, *args, **kwargs) -> None:
        """Metadata List initialization."""
        Gio.ListStore.__init__(self, item_type=Metadata)


class MetadataFile(GObject.GObject):
    """Metadata File object."""

    __gtype_name__ = "MetadataFile"

    filename = GObject.Property(type=str)
    metadata = GObject.Property(type=MetadataList)


class MetadataStore(Gio.ListStore):
    """Metadata Store object."""

    __gtype_name__ = "MetadataStore"

    def __init__(self, *args, **kwargs) -> None:
        """Metadata Store initialization."""
        Gio.ListStore.__init__(self, item_type=MetadataFile)



# SPDX-FileCopyrightText: Metadata Cleaner contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Widget crediting persons for a role."""

import re

from gi.repository import GObject, Gtk
from metadatacleaner import RESOURCE_PATH


@Gtk.Template(resource_path=RESOURCE_PATH + "/ui/CreditsRole.ui")
class CreditsRole(Gtk.Box):
    """Widget crediting persons for a role."""

    __gtype_name__ = "CreditsRole"

    role = GObject.Property(type=str)

    _persons_label: Gtk.Label = Gtk.Template.Child()

    _persons = ""

    @GObject.Property(type=str)
    def persons(self) -> str:
        """Get the persons string.

        Returns:
            str: The persons string.
        """
        return self._persons

    @persons.setter  # type: ignore
    def persons(self, value: str) -> None:
        self._persons = value
        if value:
            self._persons_label.set_label(_parse_links(value))


def _parse_links(string: str) -> str:
    lines = string.splitlines()
    parsed_lines = []
    regex_email = re.compile("<(.+)>")
    regex_url = re.compile("https?://\\S+")
    for line in lines:
        match_email = regex_email.search(line)
        match_url = regex_url.search(line)
        if match_email:
            name = line[: match_email.start()].strip()
            url = f"mailto:{match_email.group()[1:-1]}"
            parsed_lines.append(f"<a href='{url}'>{name}</a>")
        elif match_url:
            name = line[: match_url.start()].strip()
            url = match_url.group()
            parsed_lines.append(f"<a href='{url}'>{name}</a>")
        else:
            parsed_lines.append(line)
    return "\n".join(parsed_lines)
